#
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from pydantic import BaseModel, Field
from typing import Any, List, Optional

from .predictor import PredictorType

class PredictionResource (BaseModel):
    """
    Prediction resource.

    Members:
        type (str): Resource type.
        url (str): Resource URL.
    """
    type: str = Field(description="Resource type.")
    url: str = Field(description="Resource URL.")

class Prediction (BaseModel):
    """
    Prediction.

    Members:
        id (str): Prediction identifier.
        tag (str): Predictor tag.
        type (PredictorType): Prediction type.
        configuration (str): Prediction configuration token. This is only populated for `EDGE` predictions.
        resources (list): Prediction resources. This is only populated for `EDGE` predictions.
        results (list): Prediction results.
        latency (float): Prediction latency in milliseconds.
        error (str): Prediction error. This is `None` if the prediction completed successfully.
        logs (str): Prediction logs.
        created (str): Date created.
    """
    id: str = Field(description="Prediction identifier.")
    tag: str = Field(description="Predictor tag.")
    type: PredictorType = Field(description="Prediction type.")
    configuration: Optional[str] = Field(default=None, description="Prediction configuration token. This is only populated for `EDGE` predictions.")
    resources: Optional[List[PredictionResource]] = Field(default=None, description="Prediction resources. This is only populated for `EDGE` predictions.")
    results: Optional[List[Any]] = Field(default=None, description="Prediction results.")
    latency: Optional[float] = Field(default=None, description="Prediction latency in milliseconds.")
    error: Optional[str] = Field(default=None, description="Prediction error. This is `None` if the prediction completed successfully.")
    logs: Optional[str] = Field(default=None, description="Prediction logs.")
    created: str = Field(description="Date created.")


# 
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from rich import print_json
from rich.progress import Progress, SpinnerColumn, TextColumn
from pathlib import Path
from typer import Argument, Option
from typing import List

from ..function import Function
from ..types import Acceleration, AccessMode, PredictorStatus, PredictorType
from .auth import get_access_key

def retrieve_predictor (
    tag: str=Argument(..., help="Predictor tag.")
):
    fxn = Function(get_access_key())
    predictor = fxn.predictors.retrieve(tag)
    predictor = predictor.model_dump() if predictor else None
    print_json(data=predictor)

def list_predictors (
    owner: str=Option(None, help="Predictor owner. This defaults to the current user."),
    status: PredictorStatus=Option(PredictorStatus.Active, help="Predictor status. This defaults to `ACTIVE`."),
    offset: int=Option(None, help="Pagination offset."),
    count: int=Option(None, help="Pagination count.")
):
    fxn = Function(get_access_key())
    predictors = fxn.predictors.list(
        owner=owner,
        status=status,
        offset=offset,
        count=count
    )
    predictors = [predictor.model_dump() for predictor in predictors] if predictors is not None else None
    print_json(data=predictors)

def search_predictors (
    query: str=Argument(..., help="Search query."),
    offset: int=Option(None, help="Pagination offset."),
    count: int=Option(None, help="Pagination count.")
):
    fxn = Function(get_access_key())
    predictors = fxn.predictors.search(query=query, offset=offset, count=count)
    predictors = [predictor.model_dump() for predictor in predictors]
    print_json(data=predictors)

def create_predictor (
    tag: str=Argument(..., help="Predictor tag."),
    notebook: Path=Argument(..., help="Path to predictor notebook."),
    type: PredictorType=Option(None, case_sensitive=False, help="Predictor type. This defaults to `cloud`."),
    edge: bool=Option(False, "--edge", is_flag=True, help="Shorthand for `--type edge`."),
    cloud: bool=Option(False, "--cloud", is_flag=True, help="Shorthand for `--type cloud`."),
    access: AccessMode=Option(None, case_sensitive=False, help="Predictor access mode. This defaults to `private`."),
    description: str=Option(None, help="Predictor description. This must be less than 200 characters long."),
    media: Path=Option(None, help="Predictor image path."),
    acceleration: Acceleration=Option(None, case_sensitive=False, help="Cloud predictor acceleration. This defaults to `cpu`."),
    license: str=Option(None, help="Predictor license URL."),
    env: List[str]=Option([], help="Specify a predictor environment variable."),
    overwrite: bool=Option(None, "--overwrite", help="Overwrite any existing predictor with the same tag.")
):
    with Progress(
        SpinnerColumn(spinner_name="dots"),
        TextColumn("[progress.description]{task.description}"),
        transient=True
    ) as progress:
        progress.add_task(description="Analyzing Function...", total=None)
        fxn = Function(get_access_key())
        type = PredictorType.Cloud if cloud else PredictorType.Edge if edge else type
        environment = { e.split("=")[0].strip(): e.split("=")[1].strip() for e in env }
        predictor = fxn.predictors.create(
            tag=tag,
            notebook=notebook,
            type=type,
            access=access,
            description=description,
            media=media,
            acceleration=acceleration,
            environment=environment,
            license=license,
            overwrite=overwrite
        )
        print_json(data=predictor.model_dump())

def delete_predictor (
    tag: str=Argument(..., help="Predictor tag.")
):
    with Progress(
        SpinnerColumn(spinner_name="dots"),
        TextColumn("[progress.description]{task.description}"),
        transient=True
    ) as progress:
        progress.add_task(description="Deleting Function...", total=None)
        fxn = Function(get_access_key())
        result = fxn.predictors.delete(tag)
        print_json(data=result)

def archive_predictor (
    tag: str=Argument(..., help="Predictor tag.")
):
    fxn = Function(get_access_key())
    predictor = fxn.predictors.archive(tag)
    print_json(data=predictor.model_dump())


# 
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from rich import print_json
from typer import Argument, Option, Typer

from ..function import Function
from .auth import get_access_key

app = Typer(no_args_is_help=True)

@app.command(name="list", help="List environment variables.")
def list_envs (
    organization: str=Option(None, help="Organization username. Use this for organization environment variables."),
):
    fxn = Function(get_access_key())
    environments = fxn.environment_variables.list(organization=organization)
    environments = [env.model_dump() for env in environments]
    print_json(data=environments)

@app.command(name="create", help="Create an environment variable.")
def create_env (
    name: str=Argument(..., help="Variable name."),
    value: str=Argument(..., help="Variable value."),
    organization: str=Option(None, help="Organization username. Use this for organization environment variables."),
):
    fxn = Function(get_access_key())
    environment = fxn.environment_variables.create(name=name, value=value, organization=organization)
    print_json(data=environment.model_dump())

@app.command(name="delete", help="Delete an environment variable.")
def delete_env (
    name: str=Argument(..., help="Variable name."),
    organization: str=Option(None, help="Organization username. Use this for organization environment variables."),
):
    fxn = Function(get_access_key())
    result = fxn.environment_variables.delete(name=name, organization=organization)
    print_json(data=result)


#
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from aiohttp import ClientSession
from ctypes import byref, c_double, c_int32, create_string_buffer
from dataclasses import asdict, is_dataclass
from datetime import datetime, timezone
from filetype import guess_mime
from io import BytesIO
from json import dumps, loads
from numpy import array, float32, frombuffer, int32, ndarray
from numpy.typing import NDArray
from pathlib import Path
from PIL import Image
from platform import machine, system
from pydantic import BaseModel
from requests import get, post
from tempfile import NamedTemporaryFile
from typing import Any, AsyncIterator, Dict, List, Union
from uuid import uuid4
from urllib.parse import urlparse
from urllib.request import urlopen

from ...graph import GraphClient
from ...types import Dtype, PredictorType, Prediction, PredictionResource, Value, UploadType
from ..storage import StorageService
from .fxnc import load_fxnc, to_fxn_value, to_py_value, FXNConfigurationRef, FXNPredictorRef, FXNProfileRef, FXNStatus, FXNValueRef, FXNValueMapRef

class PredictionService:

    def __init__ (self, client: GraphClient, storage: StorageService) -> None:
        self.client = client
        self.storage = storage
        self.__fxnc = None
        self.__cache = { }

    def create (
        self,
        tag: str,
        *,
        inputs: Dict[str, Union[ndarray, str, float, int, bool, List, Dict[str, Any], Path, Image.Image, Value]] = None,
        raw_outputs: bool=False,
        return_binary_path: bool=True,
        data_url_limit: int=None,
        verbose: bool=False
    ) -> Prediction:
        """
        Create a prediction.

        Parameters:
            tag (str): Predictor tag.
            inputs (dict): Input values. This only applies to `CLOUD` predictions.
            raw_outputs (bool): Skip converting output values into Pythonic types. This only applies to `CLOUD` predictions.
            return_binary_path (bool): Write binary values to file and return a `Path` instead of returning `BytesIO` instance.
            data_url_limit (int): Return a data URL if a given output value is smaller than this size in bytes. This only applies to `CLOUD` predictions.
            verbose (bool): Use verbose logging.

        Returns:
            Prediction: Created prediction.
        """
        # Check if cached
        if tag in self.__cache:
            return self.__predict(tag=tag, predictor=self.__cache[tag], inputs=inputs)
        # Serialize inputs
        key = uuid4().hex
        values = { name: self.to_value(value, name, key=key).model_dump(mode="json") for name, value in inputs.items() } if inputs is not None else { }
        # Query
        response = post( # INCOMPLETE # Configuration token
            f"{self.client.api_url}/predict/{tag}?rawOutputs=true&dataUrlLimit={data_url_limit}",
            json=values,
            headers={
                "Authorization": f"Bearer {self.client.access_key}",
                "fxn-client": self.__get_client_id(),
                "fxn-configuration-token": ""
            }
        )
        # Check
        prediction = response.json()
        try:
            response.raise_for_status()
        except Exception as ex:
            error = prediction["errors"][0]["message"] if "errors" in prediction else str(ex)
            raise RuntimeError(error)
        # Parse prediction
        prediction = self.__parse_prediction(prediction, raw_outputs=raw_outputs, return_binary_path=return_binary_path)
        # Create edge prediction
        if prediction.type == PredictorType.Edge:
            predictor = self.__load(prediction, verbose=verbose)
            self.__cache[tag] = predictor
            prediction = self.__predict(tag=tag, predictor=predictor, inputs=inputs) if inputs is not None else prediction
        # Return
        return prediction
    
    async def stream ( # INCOMPLETE # Add edge prediction support
        self,
        tag: str,
        *,
        inputs: Dict[str, Union[float, int, str, bool, NDArray, List[Any], Dict[str, Any], Path, Image.Image, Value]] = {},
        raw_outputs: bool=False,
        return_binary_path: bool=True,
        data_url_limit: int=None,
        verbose: bool=False
    ) -> AsyncIterator[Prediction]:
        """
        Create a streaming prediction.

        NOTE: This feature is currently experimental.

        Parameters:
            tag (str): Predictor tag.
            inputs (dict): Input values. This only applies to `CLOUD` predictions.
            raw_outputs (bool): Skip converting output values into Pythonic types. This only applies to `CLOUD` predictions.
            return_binary_path (bool): Write binary values to file and return a `Path` instead of returning `BytesIO` instance.
            data_url_limit (int): Return a data URL if a given output value is smaller than this size in bytes. This only applies to `CLOUD` predictions.
            verbose (bool): Use verbose logging.

        Returns:
            Prediction: Created prediction.
        """
        # Serialize inputs
        key = uuid4().hex
        inputs = { name: self.to_value(value, name, key=key).model_dump(mode="json") for name, value in inputs.items() }
        # Request
        url = f"{self.client.api_url}/predict/{tag}?stream=true&rawOutputs=true&dataUrlLimit={data_url_limit}"
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.client.access_key}",
            "fxn-client": self.__get_client_id(),
            "fxn-configuration-token": ""
        }
        async with ClientSession(headers=headers) as session:
            async with session.post(url, data=dumps(inputs)) as response:
                async for chunk in response.content.iter_any():
                    prediction = loads(chunk)
                    # Check status
                    try:
                        response.raise_for_status()
                    except Exception as ex:
                        error = prediction["errors"][0]["message"] if "errors" in prediction else str(ex)
                        raise RuntimeError(error)
                    # Parse prediction
                    prediction = self.__parse_prediction(prediction, raw_outputs=raw_outputs, return_binary_path=return_binary_path)
                    # Create edge prediction
                    if prediction.type == PredictorType.Edge:
                        predictor = self.__load(prediction, verbose=verbose)
                        self.__cache[tag] = predictor
                        prediction = self.__predict(tag=tag, predictor=predictor, inputs=inputs) if inputs is not None else prediction
                    # Yield
                    yield prediction

    def to_object (
        self,
        value: Value,
        return_binary_path: bool=True
    ) -> Union[str, float, int, bool, NDArray, list, dict, Image.Image, BytesIO, Path]:
        """
        Convert a Function value to a plain object.

        Parameters:
            return_binary_path (str): Write binary values to file and return a `Path` instead of returning `BytesIO` instance.

        Returns:
            str | float | int | bool | list | dict | ndarray | Image.Image | BytesIO | Path: Plain objectt.
        """
        # Null
        if value.type == Dtype.null:
            return None
        # Download
        buffer = self.__download_value_data(value.data)
        # Array
        if value.type in [
            Dtype.int8, Dtype.int16, Dtype.int32, Dtype.int64,
            Dtype.uint8, Dtype.uint16, Dtype.uint32, Dtype.uint64,
            Dtype.float16, Dtype.float32, Dtype.float64, Dtype.bool
        ]:
            assert value.shape is not None, "Array value must have a shape specified"
            array = frombuffer(buffer.getbuffer(), dtype=value.type).reshape(value.shape)
            return array if len(value.shape) > 0 else array.item()
        # String
        if value.type == Dtype.string:
            return buffer.getvalue().decode("utf-8")
        # List
        if value.type == Dtype.list:
            return loads(buffer.getvalue().decode("utf-8"))
        # Dict
        if value.type == Dtype.dict:
            return loads(buffer.getvalue().decode("utf-8"))
        # Image
        if value.type == Dtype.image:
            return Image.open(buffer)
        # Binary
        if return_binary_path:
            with NamedTemporaryFile(mode="wb", delete=False) as f:
                f.write(buffer.getbuffer())
            return Path(f.name)
        # Return
        return buffer

    def to_value (
        self,
        object: Union[str, float, int, bool, ndarray, List[Any], Dict[str, any], Path, Image.Image],
        name: str,
        min_upload_size: int=4096,
        key: str=None
    ) -> Value:
        """
        Convert a plain object to a Function value.

        Parameters:
            object (str | float | int | bool | ndarray | list | dict | dataclass | Path | PIL.Image): Input object.
            name (str): Value name.
            min_upload_size (int): Values larger than this size in bytes will be uploaded.

        Returns:
            Value: Function value.
        """
        object = self.__try_ensure_serializable(object)
        # None
        if object is None:
            return Value(data=None, type=Dtype.null)
        # Value
        if isinstance(object, Value):
            return object
        # Array
        if isinstance(object, ndarray):
            buffer = BytesIO(object.tobytes())
            data = self.storage.upload(buffer, type=UploadType.Value, name=name, data_url_limit=min_upload_size, key=key)
            return Value(data=data, type=object.dtype.name, shape=list(object.shape))
        # String
        if isinstance(object, str):
            buffer = BytesIO(object.encode())
            data = self.storage.upload(buffer, type=UploadType.Value, name=name, data_url_limit=min_upload_size, key=key)
            return Value(data=data, type=Dtype.string)
        # Float
        if isinstance(object, float):
            object = array(object, dtype=float32)
            return self.to_value(object, name, min_upload_size=min_upload_size, key=key)
        # Boolean
        if isinstance(object, bool):
            object = array(object, dtype=bool)
            return self.to_value(object, name, min_upload_size=min_upload_size, key=key)
        # Integer
        if isinstance(object, int):
            object = array(object, dtype=int32)
            return self.to_value(object, name, min_upload_size=min_upload_size, key=key)
        # List
        if isinstance(object, list):
            buffer = BytesIO(dumps(object).encode())
            data = self.storage.upload(buffer, type=UploadType.Value, name=name, data_url_limit=min_upload_size, key=key)
            return Value(data=data, type=Dtype.list)
        # Dict
        if isinstance(object, dict):
            buffer = BytesIO(dumps(object).encode())
            data = self.storage.upload(buffer, type=UploadType.Value, name=name, data_url_limit=min_upload_size, key=key)
            return Value(data=data, type=Dtype.dict)
        # Image
        if isinstance(object, Image.Image):
            buffer = BytesIO()
            format = "PNG" if object.mode == "RGBA" else "JPEG"
            object.save(buffer, format=format)
            data = self.storage.upload(buffer, type=UploadType.Value, name=name, data_url_limit=min_upload_size, key=key)
            return Value(data=data, type=Dtype.image)
        # Binary
        if isinstance(object, BytesIO):
            data = self.storage.upload(object, type=UploadType.Value, name=name, data_url_limit=min_upload_size, key=key)
            dtype = self.__get_data_dtype(object)
            return Value(data=data, type=dtype)
        # Path
        if isinstance(object, Path):
            assert object.exists(), "Value does not exist at the given path"
            assert object.is_file(), "Value path must point to a file, not a directory"
            object = object.expanduser().resolve()
            data = self.storage.upload(object, type=UploadType.Value, name=name, data_url_limit=min_upload_size, key=key)
            dtype = self.__get_data_dtype(object)
            return Value(data=data, type=dtype)
        # Unsupported
        raise RuntimeError(f"Cannot create Function value '{name}' for object {object} of type {type(object)}")

    def __load (self, prediction: Prediction, *, verbose: bool=False):
        # Load fxnc
        if self.__fxnc is None:
            fxnc_resource = next(x for x in prediction.resources if x.type == "fxn")
            fxnc_path = self.__get_resource_path(fxnc_resource)
            self.__fxnc = load_fxnc(fxnc_path)
        # Load predictor
        fxnc = self.__fxnc
        configuration = FXNConfigurationRef()
        try:
            # Create configuration
            status = fxnc.FXNConfigurationCreate(byref(configuration))
            assert status.value == FXNStatus.OK, f"Failed to create prediction configuration for tag {prediction.tag} with status: {status.value}"
            # Populate
            status = fxnc.FXNConfigurationSetToken(configuration, prediction.configuration.encode())
            assert status.value == FXNStatus.OK, f"Failed to set prediction configuration token for tag {prediction.tag} with status: {status.value}"
            # Add resources
            for resource in prediction.resources:
                if resource.type == "fxn":
                    continue
                path = self.__get_resource_path(resource, verbose=verbose)
                status = fxnc.FXNConfigurationAddResource(configuration, resource.type.encode(), str(path).encode())
                assert status.value == FXNStatus.OK, f"Failed to set prediction configuration resource with type {resource.type} for tag {prediction.tag} with status: {status.value}"
            # Create predictor
            predictor = FXNPredictorRef()
            status = fxnc.FXNPredictorCreate(prediction.tag.encode(), configuration, byref(predictor))
            assert status.value == FXNStatus.OK, f"Failed to create prediction for tag {prediction.tag} with status: {status.value}"
            # Return
            return predictor
        finally:
            fxnc.FXNConfigurationRelease(configuration)

    def __predict (self, *, tag: str, predictor, inputs: Dict[str, Any]) -> Prediction: # DEPLOY
        fxnc = self.__fxnc
        predictor = self.__cache[tag]
        profile = FXNProfileRef()
        input_map = FXNValueMapRef()
        output_map = FXNValueMapRef()
        try:
            # Create input map
            status = fxnc.FXNValueMapCreate(byref(input_map))
            assert status.value == FXNStatus.OK, f"Failed to create prediction for tag {tag} because input values could not be provided to the predictor with status: {status.value}"
            # Marshal inputs
            for name, value in inputs.items():
                fxnc.FXNValueMapSetValue(input_map, name.encode(), to_fxn_value(fxnc, value, copy=False))
            # Predict
            status = fxnc.FXNPredictorPredict(predictor, input_map, byref(profile), byref(output_map))
            assert status.value == FXNStatus.OK, f"Failed to create prediction for tag {tag} with status: {status.value}"
            # Marshal profile
            id = create_string_buffer(256)
            error = create_string_buffer(2048)
            latency = c_double()
            log_length = c_int32()
            fxnc.FXNProfileGetID(profile, id, len(id))
            fxnc.FXNProfileGetLatency(profile, byref(latency))
            fxnc.FXNProfileGetError(profile, error, len(error))
            fxnc.FXNProfileGetLogLength(profile, byref(log_length))
            id = id.value.decode("utf-8")
            latency = latency.value
            error = error.value.decode("utf-8")
            # Marshal logs
            logs = create_string_buffer(log_length.value + 1)
            fxnc.FXNProfileGetLogs(profile, logs, len(logs))
            logs = logs.value.decode("utf-8")
            # Marshal outputs
            results = []
            output_count = c_int32()
            fxnc.FXNValueMapGetSize(output_map, byref(output_count))
            for idx in range(output_count.value):
                name = create_string_buffer(256)
                value = FXNValueRef()
                fxnc.FXNValueMapGetKey(output_map, idx, name, len(name))
                fxnc.FXNValueMapGetValue(output_map, name, byref(value))
                name = name.value.decode("utf-8")
                value = to_py_value(fxnc, value)
                results.append(value)
            # Return
            return Prediction(
                id=id,
                tag=tag,
                type=PredictorType.Edge,
                results=results if not error else None,
                latency=latency,
                error=error,
                logs=logs,
                created=datetime.now(timezone.utc).isoformat()
            )
        finally:
            fxnc.FXNProfileRelease(profile)
            #fxnc.FXNValueMapRelease(input_map)
            fxnc.FXNValueMapRelease(output_map)

    def __parse_prediction (self, data: Dict[str, Any], *, raw_outputs: bool, return_binary_path: bool) -> Prediction:
        prediction = Prediction(**data)
        prediction.results = [Value(**value) for value in prediction.results] if prediction.results is not None else None
        prediction.results = [self.to_object(value, return_binary_path=return_binary_path) for value in prediction.results] if prediction.results is not None and not raw_outputs else prediction.results
        return prediction

    def __get_data_dtype (self, data: Union[Path, BytesIO]) -> Dtype:
        mime = guess_mime(str(data) if isinstance(data, Path) else data)
        if not mime:
            return Dtype.binary
        if mime.startswith("image"):
            return Dtype.image
        if mime.startswith("video"):
            return Dtype.video
        if mime.startswith("audio"):
            return Dtype.audio
        if isinstance(data, Path) and data.suffix in [".obj", ".gltf", ".glb", ".fbx", ".usd", ".usdz", ".blend"]:
            return Dtype._3d
        return Dtype.binary

    def __download_value_data (self, url: str) -> BytesIO:
        if url.startswith("data:"):
            with urlopen(url) as response:
                return BytesIO(response.read())
        response = get(url)
        result = BytesIO(response.content)
        return result
    
    def __get_resource_path (self, resource: PredictionResource, *, verbose: bool=False) -> Path: # INCOMPLETE # Verbose
        cache_dir = Path.home() / ".fxn" / "cache"
        cache_dir.mkdir(exist_ok=True)
        res_name = Path(urlparse(resource.url).path).name
        res_path = cache_dir / res_name
        if res_path.exists():
            return res_path
        req = get(resource.url)
        req.raise_for_status()
        with open(res_path, "wb") as f:
            f.write(req.content)
        return res_path
    
    @classmethod
    def __try_ensure_serializable (cls, object: Any) -> Any:
        if object is None:
            return object
        if isinstance(object, Value): # passthrough
            return object
        if isinstance(object, list):
            return [cls.__try_ensure_serializable(x) for x in object]
        if is_dataclass(object) and not isinstance(object, type):
            return asdict(object)
        if isinstance(object, BaseModel):
            return object.model_dump(mode="json", by_alias=True)
        return object
    
    @classmethod
    def __get_client_id (self) -> str:
        id = system()
        if id == "Darwin":
            return f"macos:{machine()}"
        if id == "Linux":
            return f"linux:{machine()}"
        if id == "Windows":
            return f"windows:{machine()}"
        raise RuntimeError(f"Function cannot make predictions on the {id} platform")
    

PREDICTION_FIELDS = f"""
id
tag
type
configuration
resources {{
    id
    type
    url
}}
results {{
    data
    type
    shape
}}
latency
error
logs
created
"""


#
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from typing import List

from ..graph import GraphClient
from ..types import EnvironmentVariable

class EnvironmentVariableService:

    def __init__ (self, client: GraphClient) -> None:
        self.client = client
        self.__value = "xxxxxxxx"

    def list (self, organization: str=None) -> List[EnvironmentVariable]:
        """
        List the current user's environment variables.

        Note that the variable values can only viewed at https://fxn.ai.

        Parameters:
            organization (str): Organization username.

        Returns:
            list: User environment variables.
        """
        # Query
        response = self.client.query(f"""
            query ($input: UserInput) {{
                user (input: $input) {{
                    ... on User {{
                        environmentVariables {{
                            {ENVIRONMENT_VARIABLE_FIELDS}
                        }}
                    }}
                    ... on Organization {{
                        environmentVariables {{
                            {ENVIRONMENT_VARIABLE_FIELDS}
                        }}
                    }}
                }}
            }}
            """,
            { "input": { "username": organization } if organization is not None else None }
        )
        # Create envs
        assert response["user"] is not None, "Failed to list environment variables because user could not be found. Check that you are authenticated."
        environments = response["user"]["environmentVariables"]
        environments = [EnvironmentVariable(**env, value=self.__value) for env in environments]
        # Return
        return environments
    
    def create (self, name: str, value: str, organization: str=None) -> EnvironmentVariable:
        """
        Create an environment variable.

        This environment variable will apply to all predictors you create.

        Parameters:
            name (str): Variable name.
            value (str): Variable value.
            organization (str): Organization username. Use this for organization environment variables.

        Returns:
            EnvironmentVariable: Created environment variable.
        """
        # Query
        response = self.client.query(f"""
            mutation ($input: CreateEnvironmentVariableInput!) {{
                environment: createEnvironmentVariable (input: $input) {{
                    {ENVIRONMENT_VARIABLE_FIELDS}
                }}
            }}
            """,
            { "input": { "name": name, "value": value, "organization": organization } }
        )
        # Create env
        environment = response["environment"]
        environment = EnvironmentVariable(**environment, value=self.__value)
        # Return
        return environment
    
    def delete (self, name: str, organization: str=None) -> bool:
        """
        Delete an environment variable.

        Parameters:
            name (str): Variable name.
            organization (str): Organization username. Use this for organization environment variables.
            access_key (str): Function access key.
        
        Returns:
            bool: Whether the environment variable was successfully deleted.
        """
        # Query
        response = self.client.query(f"""
            mutation ($input: DeleteEnvironmentVariableInput!) {{
                result: deleteEnvironmentVariable (input: $input)
            }}
            """,
            { "input": { "name": name, "organization": organization } }
        )
        # Return
        return response["result"]

    
ENVIRONMENT_VARIABLE_FIELDS = f"""
name
"""


# 
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from pathlib import Path
from rich import print, print_json
from typer import Argument, Typer

from ..function import Function

app = Typer(no_args_is_help=True)

@app.command(name="login", help="Login to Function.")
def login (
    access_key: str=Argument(..., help="Function access key.", envvar="FXN_ACCESS_KEY")
):
    fxn = Function(access_key)
    user = fxn.users.retrieve()
    user = user.model_dump() if user else None
    _set_access_key(access_key if user is not None else None)
    print_json(data=user)

@app.command(name="status", help="Get current authentication status.")
def auth_status ():
    fxn = Function(get_access_key())
    user = fxn.users.retrieve()
    user = user.model_dump() if user else None
    print_json(data=user)

@app.command(name="logout", help="Logout from Function.")
def logout ():
    _set_access_key(None)
    print("Successfully logged out of Function")

def get_access_key () -> str:
    """
    Get the CLI access key.

    Returns:
        str: CLI access key.
    """
    credentials_path = Path.home() / ".fxn" / "credentials"
    if not credentials_path.exists():
        return None
    with open(credentials_path) as f:
        return f.read()

def _set_access_key (key: str):
    """
    Set the CLI access key.

    Parameters:
        key (str); CLI access key.
    """
    credentials_path = Path.home() / ".fxn" / "credentials"
    credentials_path.parent.mkdir(parents=True, exist_ok=True)
    if key:
        with open(credentials_path, "w") as f:
            f.write(key)
    elif credentials_path.exists():
        credentials_path.unlink()


#
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from __future__ import annotations
from pydantic import BaseModel, Field

class Tag (BaseModel):
    """
    Predictor tag.

    Members:
        username (str): Predictor owner username.
        name (str): Predictor name.
    """
    username: str = Field(description="Predictor owner username.")
    name: str = Field(description="Predictor name.")

    def from_str (cls, tag: str) -> Tag:
        """
        Parse a predictor tag from a string.
        """
        username, name = tag.lower()[1:].split("/")
        return Tag(username=username, name=name)

    def __str__ (self):
        return f"@{self.username}/{self.name}"


# 
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from asyncio import run as run_async
from io import BytesIO
from numpy import ndarray
from pathlib import Path, PurePath
from PIL import Image
from rich import print_json
from rich.progress import Progress, SpinnerColumn, TextColumn
from tempfile import mkstemp
from typer import Argument, Context, Option
from typing import Any, Dict

from ..function import Function
from .auth import get_access_key

def predict (
    tag: str = Argument(..., help="Predictor tag."),
    raw_outputs: bool = Option(False, "--raw-outputs", help="Output raw Function values instead of converting into plain Python values."),
    context: Context = 0
):
    run_async(_predict_async(tag, context=context, raw_outputs=raw_outputs))

async def _predict_async (tag: str, context: Context, raw_outputs: bool):
    with Progress(
        SpinnerColumn(spinner_name="dots"),
        TextColumn("[progress.description]{task.description}"),
        transient=True
    ) as progress:
        progress.add_task(description="Running Function...", total=None)
        # Parse inputs
        inputs = { context.args[i].replace("-", ""): _parse_value(context.args[i+1]) for i in range(0, len(context.args), 2) }
        # Stream
        fxn = Function(get_access_key())
        async for prediction in fxn.predictions.stream(tag, inputs=inputs, raw_outputs=raw_outputs, return_binary_path=True, verbose=True):
            # Parse results
            images = [value for value in prediction.results or [] if isinstance(value, Image.Image)]
            prediction.results = [_serialize_value(value) for value in prediction.results] if prediction.results is not None else None
            # Print
            print_json(data=prediction.model_dump())
            # Show images
            for image in images:
                image.show()

def _parse_value (value: str):
    """
    Parse a value from a CLI argument.

    Parameters:
        value (str): CLI input argument.

    Returns:
        bool | int | float | str | Path: Parsed value.
    """
    # Boolean
    if value == "true":
        return True
    if value == "false":
        return False
    # Integer
    try:
        return int(value)
    except ValueError:
        pass
    # Float
    try:
        return float(value)
    except ValueError:
        pass
    # File
    if value.startswith("@"):
        return Path(value[1:])
    # String
    return value
    
def _serialize_value (value):
    # Convert ndarray to list
    if isinstance(value, ndarray):
        return value.tolist()
    # Write image
    if isinstance(value, Image.Image):
        _, path = mkstemp(suffix=".png" if value.mode == "RGBA" else ".jpg")
        value.save(path)
        return path
    # Serialize `BytesIO`
    if isinstance(value, BytesIO):
        return str(value)
    # Serialize `Path`
    if isinstance(value, PurePath):
        return str(value)
    # Return    
    return value

def _prediction_dict_factory (kv_pairs):
    # Check if value
    VALUE_KEYS = ["data", "type", "shape"]
    keys = [k for k, _ in kv_pairs]
    is_value = all(k in keys for k in VALUE_KEYS)
    kv_pairs = [(k, v) for k, v in kv_pairs if v is not None] if is_value else kv_pairs
    # Construct
    return dict(kv_pairs)


#
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from .user import UserService, PROFILE_FIELDS, USER_FIELDS
from .predictor import PredictorService, PREDICTOR_FIELDS
from .prediction import PredictionService, PREDICTION_FIELDS
from .environment import EnvironmentVariableService, ENVIRONMENT_VARIABLE_FIELDS
from .storage import StorageService


# 
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from enum import Enum

class UploadType (str, Enum):
    """
    Upload URL type.
    """
    Media = "MEDIA"
    Notebook = "NOTEBOOK"
    Value = "VALUE"


# 
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from pydantic import BaseModel, Field
from typing import Optional

class Profile (BaseModel):
    """
    Function user profile.

    Members:
        username (str): Username.
        email (str): User email address.
        created (str): Date created.
        name (str): User display name.
        avatar (str): User avatar URL.
        bio (str): User bio.
        website (str): User website.
        github (str): User GitHub handle.
    """
    username: str = Field(description="Username.")
    email: Optional[str] = Field(default=None, description="User email address.")
    created: Optional[str] = Field(default=None, description="Date created.")
    name: Optional[str] = Field(default=None, description="User display name.")
    avatar: Optional[str] = Field(default=None, description="User avatar URL.")
    bio: Optional[str] = Field(default=None, description="User bio.")
    website: Optional[str] = Field(default=None, description="User website.")
    github: Optional[str] = Field(default=None, description="User GitHub handle.")


# 
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from ..graph import GraphClient
from ..types import Profile

class UserService:

    def __init__ (self, client: GraphClient) -> None:
        self.client = client

    def retrieve (self, username: str=None) -> Profile:
        """
        Retrieve a user.

        Parameters:
            username (str): Username. If `None`, this will retrieve the currently authenticated user.
            access_key (str): Function access key.

        Returns:
            User: User.
        """
        # Query
        response = self.client.query(f"""
            query {"($input: UserInput)" if username else ""} {{
                user {"(input: $input)" if username else ""} {{
                    {PROFILE_FIELDS}
                    {USER_FIELDS if not username else ""}
                }}
            }}
            """,
            { "input": { "username": username } },
        )
        # Create user
        user = response["user"]
        user = Profile(**user) if user else None
        # Return
        return user

PROFILE_FIELDS = f"""
username
created
name
avatar
bio
website
github
"""

USER_FIELDS = f"""
... on User {{
    email
}}
"""


# 
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from enum import Enum
from pydantic import AliasChoices, BaseModel, Field
from typing import List, Optional, Tuple, Union

from .dtype import Dtype
from .profile import Profile
from .value import Value

class Acceleration (str, Enum):
    """
    Predictor acceleration.
    """
    CPU = "CPU"
    A40 = "A40"
    A100 = "A100"

class AccessMode (str, Enum):
    """
    Predictor access mode.
    """
    Public = "PUBLIC"
    Private = "PRIVATE"

class PredictorType (str, Enum):
    """
    Predictor type.
    """
    Cloud = "CLOUD"
    Edge = "EDGE"

class PredictorStatus (str, Enum):
    """
    Predictor status.
    """
    Provisioning = "PROVISIONING"
    Active = "ACTIVE"
    Invalid = "INVALID"
    Archived = "ARCHIVED"

class EnumerationMember (BaseModel):
    """
    Prediction parameter enumeration member.

    Members:
        name (str): Enumeration member name.
        value (str | int): Enumeration member value.
    """
    name: str = Field(description="Enumeration member name.")
    value: Union[str, int] = Field(description="Enumeration member value.")

class Parameter (BaseModel):
    """
    Predictor parameter.

    Members:
        name (str): Parameter name.
        type (Dtype): Parameter type. This is `None` if the type is unknown or unsupported by Function.
        description (str): Parameter description.
        optional (bool): Whether the parameter is optional.
        range (tuple): Parameter value range for numeric parameters.
        enumeration (list): Parameter value choices for enumeration parameters.
        default_value (Value): Parameter default value.
        value_schema (dict): Parameter JSON schema. This is only populated for `list` and `dict` parameters.
    """
    name: str = Field(description="Parameter name.")
    type: Optional[Dtype] = Field(default=None, description="Parameter type. This is `None` if the type is unknown or unsupported by Function.")
    description: Optional[str] = Field(default=None, description="Parameter description.")
    optional: Optional[bool] = Field(default=None, description="Whether the parameter is optional.")
    range: Optional[Tuple[float, float]] = Field(default=None, description="Parameter value range for numeric parameters.")
    enumeration: Optional[List[EnumerationMember]] = Field(default=None, description="Parameter value choices for enumeration parameters.")
    default_value: Optional[Value] = Field(default=None, description="Parameter default value.", serialization_alias="defaultValue", validation_alias=AliasChoices("default_value", "defaultValue"))
    value_schema: Optional[dict] = Field(default=None, description="Parameter JSON schema. This is only populated for `list` and `dict` parameters.", serialization_alias="schema", validation_alias=AliasChoices("schema", "value_schema"))

class Signature (BaseModel):
    """
    Predictor signature.

    Members:
        inputs (list): Input parameters.
        outputs (list): Output parameters.
    """
    inputs: List[Parameter] = Field(description="Input parameters.")
    outputs: List[Parameter] = Field(description="Output parameters.")

class Predictor (BaseModel):
    """
    Predictor.

    Members:
        tag (str): Predictor tag.
        owner (Profile): Predictor owner.
        name (str): Predictor name.
        type (PredictorType): Predictor type.
        status (PredictorStatus): Predictor status.
        access (AccessMode): Predictor access.
        signature (Signature): Predictor signature.
        created (str): Date created.
        description (str): Predictor description.
        card (str): Predictor card.
        media (str): Predictor media URL.
        acceleration (Acceleration): Predictor acceleration. This only applies to cloud predictors.
        license (str): Predictor license URL.
    """
    tag: str = Field(description="Predictor tag.")
    owner: Profile = Field(description="Predictor owner.")
    name: str = Field(description="Predictor name.")
    type: PredictorType = Field(description="Predictor type.")
    status: PredictorStatus = Field(description="Predictor status.")
    access: AccessMode = Field(description="Predictor access.")
    signature: Signature = Field(description="Predictor signature.")
    created: str = Field(description="Date created.")
    description: Optional[str] = Field(default=None, description="Predictor description.")
    card: Optional[str] = Field(default=None, description="Predictor card.")
    media: Optional[str] = Field(default=None, description="Predictor media URL.")
    acceleration: Optional[Acceleration] = Field(default=None, description="Predictor acceleration. This only applies to cloud predictors.")
    license: Optional[str] = Field(default=None, description="Predictor license URL.")


# 
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from os import environ

from .graph import GraphClient
from .services import EnvironmentVariableService, PredictionService, PredictorService, StorageService, UserService

class Function:
    """
    Function client.

    Members:
        client (GraphClient): Function graph API client. Do NOT use this unless you know what you are doing.
        users (UserService): Manage users.
        predictors (PredictorService): Manage predictors.
        predictions (PredictionService): Manage predictions.
        environment_variables (EnvironmentVariableService): Manage predictor environment variables.
        storage (StorageService): Upload and download files.

    Constructor:
        access_key (str): Function access key.
        api_url (str): Function API URL.
    """
    client: GraphClient
    users: UserService
    predictors: PredictorService
    predictions: PredictionService
    environment_variables: EnvironmentVariableService
    storage: StorageService

    def __init__ (self, access_key: str=None, api_url: str=None):
        access_key = access_key or environ.get("FXN_ACCESS_KEY", None)
        api_url = api_url or environ.get("FXN_API_URL", "https://api.fxn.ai")
        self.client = GraphClient(access_key, api_url)
        self.users = UserService(self.client)
        self.storage = StorageService(self.client)
        self.predictors = PredictorService(self.client, self.storage)
        self.predictions = PredictionService(self.client, self.storage)
        self.environment_variables = EnvironmentVariableService(self.client)


#
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from requests import post
from typing import Any, Dict

class GraphClient:
    """
    Function graph API client.
    """
    
    def __init__(self, access_key: str, api_url: str) -> None:
        self.access_key = access_key
        self.api_url = api_url

    def query (self, query: str, variables: dict=None) -> Dict[str, Any]:
        """
        Query the Function graph API.

        Parameters:
            query (str): Graph query.
            variables (dict): Input variables.

        Returns:
            dict: Response dictionary.
        """
        # Request
        response = post(
            f"{self.api_url}/graph",
            json={ "query": query, "variables": variables },
            headers={ "Authorization": f"Bearer {self.access_key}" } if self.access_key else { }
        )
        # Check
        payload = response.json()
        try:
            response.raise_for_status()
        except:
            raise RuntimeError(payload.get("error"))
        # Check error
        if "errors" in payload:
            raise RuntimeError(payload["errors"][0]["message"])
        # Return
        return payload["data"]


#
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from ctypes import byref, cast, c_bool, c_char_p, c_double, c_int, c_int32, c_uint8, c_void_p, string_at, CDLL, POINTER, Structure, _Pointer
from io import BytesIO
from json import dumps, loads
from numpy import array, dtype, int32, ndarray, zeros
from numpy.ctypeslib import as_array, as_ctypes_type
from numpy.typing import NDArray
from pathlib import Path
from PIL import Image
from typing import Any, Dict, List, Union

# https://github.com/fxnai/fxnc

class FXNStatus(c_int):
    OK = 0
    ERROR_INVALID_ARGUMENT = 1
    ERROR_INVALID_OPERATION = 2
    ERROR_NOT_IMPLEMENTED = 3

class FXNDtype(c_int):
    NULL = 0
    FLOAT16 = 1
    FLOAT32 = 2
    FLOAT64 = 3
    INT8 = 4
    INT16 = 5
    INT32 = 6
    INT64 = 7
    UINT8 = 8
    UINT16 = 9
    UINT32 = 10
    UINT64 = 11
    BOOL = 12
    STRING = 13
    LIST = 14
    DICT = 15
    IMAGE = 16
    BINARY = 17

class FXNValueFlags(c_int):
    NONE = 0
    COPY_DATA = 1

class FXNAcceleration(c_int):
    FXN_ACCELERATION_DEFAULT = 0
    FXN_ACCELERATION_CPU = 1 << 0
    FXN_ACCELERATION_GPU = 1 << 1
    FXN_ACCELERATION_NPU = 1 << 2

class FXNValue(Structure): pass
class FXNValueMap(Structure): pass
class FXNConfiguration(Structure): pass
class FXNProfile(Structure): pass
class FXNPredictor(Structure): pass

FXNValueRef = POINTER(FXNValue)
FXNValueMapRef = POINTER(FXNValueMap)
FXNConfigurationRef = POINTER(FXNConfiguration)
FXNProfileRef = POINTER(FXNProfile)
FXNPredictorRef = POINTER(FXNPredictor)

def load_fxnc (path: Path) -> CDLL:
    # Open
    fxnc = CDLL(str(path))
    # FXNValueRelease
    fxnc.FXNValueRelease.argtypes = [FXNValueRef]
    fxnc.FXNValueRelease.restype = FXNStatus
    # FXNValueGetData
    fxnc.FXNValueGetData.argtypes = [FXNValueRef, POINTER(c_void_p)]
    fxnc.FXNValueGetData.restype = FXNStatus
    # FXNValueGetType
    fxnc.FXNValueGetType.argtypes = [FXNValueRef, POINTER(FXNDtype)]
    fxnc.FXNValueGetType.restype = FXNStatus
    # FXNValueGetDimensions
    fxnc.FXNValueGetDimensions.argtypes = [FXNValueRef, POINTER(c_int32)]
    fxnc.FXNValueGetDimensions.restype = FXNStatus
    # FXNValueGetShape
    fxnc.FXNValueGetShape.argtypes = [FXNValueRef, POINTER(c_int32), c_int32]
    fxnc.FXNValueGetShape.restype = FXNStatus
    # FXNValueCreateArray
    fxnc.FXNValueCreateArray.argtypes = [c_void_p, POINTER(c_int32), c_int32, FXNDtype, FXNValueFlags, POINTER(FXNValueRef)]
    fxnc.FXNValueCreateArray.restype = FXNStatus
    # FXNValueCreateString
    fxnc.FXNValueCreateString.argtypes = [c_char_p, POINTER(FXNValueRef)]
    fxnc.FXNValueCreateString.restype = FXNStatus
    # FXNValueCreateList
    fxnc.FXNValueCreateList.argtypes = [c_char_p, POINTER(FXNValueRef)]
    fxnc.FXNValueCreateList.restype = FXNStatus
    # FXNValueCreateDict
    fxnc.FXNValueCreateDict.argtypes = [c_char_p, POINTER(FXNValueRef)]
    fxnc.FXNValueCreateDict.restype = FXNStatus
    # FXNValueCreateImage
    fxnc.FXNValueCreateImage.argtypes = [POINTER(c_uint8), c_int32, c_int32, c_int32, FXNValueFlags, POINTER(FXNValueRef)]
    fxnc.FXNValueCreateImage.restype = FXNStatus
    # FXNValueMapCreate
    fxnc.FXNValueMapCreate.argtypes = [POINTER(FXNValueMapRef)]
    fxnc.FXNValueMapCreate.restype = FXNStatus
    # FXNValueMapRelease
    fxnc.FXNValueMapRelease.argtypes = [FXNValueMapRef]
    fxnc.FXNValueMapRelease.restype = FXNStatus
    # FXNValueMapGetSize
    fxnc.FXNValueMapGetSize.argtypes = [FXNValueMapRef, POINTER(c_int32)]
    fxnc.FXNValueMapGetSize.restype = FXNStatus
    # FXNValueMapGetKey
    fxnc.FXNValueMapGetKey.argtypes = [FXNValueMapRef, c_int32, c_char_p, c_int32]
    fxnc.FXNValueMapGetKey.restype = FXNStatus
    # FXNValueMapGetValue
    fxnc.FXNValueMapGetValue.argtypes = [FXNValueMapRef, c_char_p, POINTER(FXNValueRef)]
    fxnc.FXNValueMapGetValue.restype = FXNStatus
    # FXNValueMapSetValue
    fxnc.FXNValueMapSetValue.argtypes = [FXNValueMapRef, c_char_p, FXNValueRef]
    fxnc.FXNValueMapSetValue.restype = FXNStatus
    # FXNConfigurationGetUniqueID
    fxnc.FXNConfigurationGetUniqueID.argtypes = [c_char_p, c_int32]
    fxnc.FXNConfigurationGetUniqueID.restype = FXNStatus
    # FXNConfigurationCreate
    fxnc.FXNConfigurationCreate.argtypes = [POINTER(FXNConfigurationRef)]
    fxnc.FXNConfigurationCreate.restype = FXNStatus
    # FXNConfigurationRelease
    fxnc.FXNConfigurationRelease.argtypes = [FXNConfigurationRef]
    fxnc.FXNConfigurationRelease.restype = FXNStatus
    # FXNConfigurationGetToken
    fxnc.FXNConfigurationGetToken.argtypes = [FXNConfigurationRef, c_char_p, c_int32]
    fxnc.FXNConfigurationGetToken.restype = FXNStatus
    # FXNConfigurationSetToken
    fxnc.FXNConfigurationSetToken.argtypes = [FXNConfigurationRef, c_char_p]
    fxnc.FXNConfigurationSetToken.restype = FXNStatus
    # FXNConfigurationAddResource
    fxnc.FXNConfigurationAddResource.argtypes = [FXNConfigurationRef, c_char_p, c_char_p]
    fxnc.FXNConfigurationAddResource.restype = FXNStatus
    # FXNConfigurationGetAcceleration
    fxnc.FXNConfigurationGetAcceleration.argtypes = [FXNConfigurationRef, POINTER(FXNAcceleration)]
    fxnc.FXNConfigurationGetAcceleration.restype = FXNStatus
    # FXNConfigurationSetAcceleration
    fxnc.FXNConfigurationSetAcceleration.argtypes = [FXNConfigurationRef, FXNAcceleration]
    fxnc.FXNConfigurationSetAcceleration.restype = FXNStatus
    # FXNConfigurationGetDevice
    fxnc.FXNConfigurationGetDevice.argtypes = [FXNConfigurationRef, POINTER(c_void_p)]
    fxnc.FXNConfigurationGetDevice.restype = FXNStatus
    # FXNConfigurationSetDevice
    fxnc.FXNConfigurationSetDevice.argtypes = [FXNConfigurationRef, c_void_p]
    fxnc.FXNConfigurationSetDevice.restype = FXNStatus
    # FXNProfileRelease
    fxnc.FXNProfileRelease.argtypes = [FXNProfileRef]
    fxnc.FXNProfileRelease.restype = FXNStatus
    # FXNProfileGetID
    fxnc.FXNProfileGetID.argtypes = [FXNProfileRef, c_char_p, c_int32]
    fxnc.FXNProfileGetID.restype = FXNStatus
    # FXNProfileGetLatency
    fxnc.FXNProfileGetLatency.argtypes = [FXNProfileRef, POINTER(c_double)]
    fxnc.FXNProfileGetLatency.restype = FXNStatus
    # FXNProfileGetError
    fxnc.FXNProfileGetError.argtypes = [FXNProfileRef, c_char_p, c_int32]
    fxnc.FXNProfileGetError.restype = FXNStatus
    # FXNProfileGetLogs
    fxnc.FXNProfileGetLogs.argtypes = [FXNProfileRef, c_char_p, c_int32]
    fxnc.FXNProfileGetLogs.restype = FXNStatus
    # FXNProfileGetLogLength
    fxnc.FXNProfileGetLogLength.argtypes = [FXNProfileRef, POINTER(c_int32)]
    fxnc.FXNProfileGetLogLength.restype = FXNStatus
    # FXNPredictorCreate
    fxnc.FXNPredictorCreate.argtypes = [c_char_p, FXNConfigurationRef, POINTER(FXNPredictorRef)]
    fxnc.FXNPredictorCreate.restype = FXNStatus
    # FXNPredictorRelease
    fxnc.FXNPredictorRelease.argtypes = [FXNPredictorRef]
    fxnc.FXNPredictorRelease.restype = FXNStatus
    # FXNPredictorPredict
    fxnc.FXNPredictorPredict.argtypes = [FXNPredictorRef, FXNValueMapRef, POINTER(FXNProfileRef), POINTER(FXNValueMapRef)]
    fxnc.FXNPredictorPredict.restype = FXNStatus
    # FXNGetVersion
    fxnc.FXNGetVersion.argtypes = []
    fxnc.FXNGetVersion.restype = c_char_p
    # Return
    return fxnc

def to_fxn_value ( # DEPLOY
    fxnc: CDLL,
    value: Union[float, int, bool, str, NDArray, List[Any], Dict[str, Any], Image.Image, bytes, bytearray, memoryview, BytesIO, None],
    *,
    copy: bool=False
) -> type[FXNValueRef]:
    result = FXNValueRef()
    if result is None:
        fxnc.FXNValueCreateNull(byref(result))
    elif isinstance(value, bool):
        return to_fxn_value(fxnc, array(value, dtype="bool"))
    elif isinstance(value, int):
        return to_fxn_value(fxnc, array(value, dtype="int32"))
    elif isinstance(value, float):
        return to_fxn_value(fxnc, array(value, dtype="float32"))
    elif isinstance(value, ndarray):
        dtype = _NP_TO_FXN_DTYPE.get(value.dtype)
        assert dtype is not None, f"Failed to convert numpy array to Function value because array data type is not supported: {value.dtype}"
        fxnc.FXNValueCreateArray(
            value.ctypes.data_as(c_void_p),
            value.ctypes.shape_as(c_int32),
            len(value.shape),
            dtype,
            FXNValueFlags.COPY_DATA if copy else FXNValueFlags.NONE,
            byref(result)
        )
    elif isinstance(value, str):
        fxnc.FXNValueCreateString(value.encode(), byref(result))
    elif isinstance(value, list):
        fxnc.FXNValueCreateList(dumps(value).encode(), byref(result))
    elif isinstance(value, dict):
        fxnc.FXNValueCreateList(dumps(value).encode(), byref(result))
    elif isinstance(value, Image.Image):
        value = array(value)
        fxnc.FXNValueCreateImage(
            value.ctypes.data_as(c_void_p),
            value.shape[1],
            value.shape[0],
            value.shape[2],
            FXNValueFlags.COPY_DATA,
            byref(result)
        )
    elif isinstance(value, (bytes, bytearray, memoryview, BytesIO)):
        view = memoryview(value.getvalue() if isinstance(value, BytesIO) else value) if not isinstance(value, memoryview) else value
        buffer = (c_uint8 * len(view)).from_buffer(view)
        fxnc.FXNValueCreateBinary(
            buffer,
            len(view),
            FXNValueFlags.COPY_DATA if copy else FXNValueFlags.NONE,
            byref(result)
        )
    else:
        raise RuntimeError(f"Failed to convert Python value to Function value because Python value has an unsupported type: {type(value)}")
    return result

def to_py_value ( # DEPLOY
    fxnc: CDLL,
    value: type[FXNValueRef]
) -> Union[float, int, bool, str, NDArray, List[Any], Dict[str, Any], Image.Image, BytesIO, None]:
    # Type
    dtype = FXNDtype()
    fxnc.FXNValueGetType(value, byref(dtype))
    dtype = dtype.value
    # Get data
    data = c_void_p()
    fxnc.FXNValueGetData(value, byref(data))
    # Get shape
    dims = c_int32()
    fxnc.FXNValueGetDimensions(value, byref(dims))
    shape = zeros(dims.value, dtype=int32)
    fxnc.FXNValueGetShape(value, shape.ctypes.data_as(POINTER(c_int32)), dims)
    # Switch
    if dtype == FXNDtype.NULL:
        return None
    elif dtype in _FXN_TO_NP_DTYPE:
        dtype_c = as_ctypes_type(_FXN_TO_NP_DTYPE[dtype])
        tensor = as_array(cast(data, POINTER(dtype_c)), shape)
        return tensor.item() if len(tensor.shape) == 0 else tensor.copy()
    elif dtype == FXNDtype.STRING:
        return cast(data, c_char_p).value.decode()
    elif dtype == FXNDtype.LIST:
        return loads(cast(data, c_char_p).value.decode())
    elif dtype == FXNDtype.DICT:
        return loads(cast(data, c_char_p).value.decode())
    elif dtype == FXNDtype.IMAGE:
        pixel_buffer = as_array(cast(data, POINTER(c_uint8)), shape)
        return Image.fromarray(pixel_buffer)
    elif dtype == FXNDtype.BINARY:
        return BytesIO(string_at(data, shape[0]))
    else:
        raise RuntimeError(f"Failed to convert Function value to Python value because Function value has unsupported type: {dtype}")

_FXN_TO_NP_DTYPE = {
    FXNDtype.FLOAT16:   dtype("float16"),
    FXNDtype.FLOAT32:   dtype("float32"),
    FXNDtype.FLOAT64:   dtype("float64"),
    FXNDtype.INT8:      dtype("int8"),
    FXNDtype.INT16:     dtype("int16"),
    FXNDtype.INT32:     dtype("int32"),
    FXNDtype.INT64:     dtype("int64"),
    FXNDtype.UINT8:     dtype("uint8"),
    FXNDtype.UINT16:    dtype("uint16"),
    FXNDtype.UINT32:    dtype("uint32"),
    FXNDtype.UINT64:    dtype("uint64"),
    FXNDtype.BOOL:      dtype("bool"),
}

_NP_TO_FXN_DTYPE = { value: key for key, value in _FXN_TO_NP_DTYPE.items() }


#
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from pydantic import BaseModel, Field
from typing import List, Optional, Union

from .dtype import Dtype

class Value (BaseModel):
    """
    Prediction value.

    Members:
        data (str): Value URL. This can be a web URL or a data URL.
        type (Dtype): Value data type.
        shape (list): Value shape. This is `None` if shape information is not available or applicable.
    """
    data: Union[str, None] = Field(description="Value URL. This can be a web URL or a data URL.")
    type: Dtype = Field(description="Value data type.")
    shape: Optional[List[int]] = Field(default=None, description="Value shape. This is `None` if shape information is not available or applicable.")